from pydantic import SecretStr
from fakts.grants.remote import FaktsEndpoint
from .types import Token
from pydantic import BaseModel


class StaticDemander(BaseModel):
    """Static Grant

    A static demander is a remote grant that has a static token. This token can
    for example have been retrieved from a configuration file beforehand and uniquely
    identifies the application on the fakts server. When using the static grant make
    sure that the token is not shared with other applications. As they can then mimik
    your application.

    Attention: If you are using the static grant, make sure that the token is not
    shared with other applications. As they can then mimik your application, especially
    when this static token maps to an client-credentials (user) application on the fakts
    server, as this application will then be able to access the data of the user that
    granted the application in the first place.

    """

    token: SecretStr
    """ The token (secret) that uniquely identifies this application on the fakts server."""

    async def ademand(self, endpoint: FaktsEndpoint, request) -> Token:
        return self.token.get_secret_value()



import json
from typing import runtime_checkable, Protocol
from herre.grants.stored_login import (
    UserStore,
    StoredUser,
)
from qtpy import QtWidgets, QtCore
from fakts.grants.remote.types import FaktsEndpoint
from koil.qt import qt_to_async


class ShouldWeSaveThisAsDefault(QtWidgets.QDialog):
    def __init__(self, stored: FaktsEndpoint = None, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle(f"Connected to {stored.name}")

        self.qlabel = QtWidgets.QLabel(
            f"Do you want to save this endpoint as the default endpoint?"
        )

        vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(vlayout)

        vlayout.addWidget(self.qlabel)

        hlayout = QtWidgets.QHBoxLayout()
        vlayout.addLayout(hlayout)

        self.yes_button = QtWidgets.QPushButton("Yes")
        self.no_button = QtWidgets.QPushButton("No")

        self.yes_button.clicked.connect(self.on_yes)
        self.no_button.clicked.connect(self.on_no)

        self.stored = stored

        hlayout.addWidget(self.yes_button)
        hlayout.addWidget(self.no_button)

    def on_yes(self):
        self.accept()

    def on_no(self):
        self.reject()


class AutoSaveEndpointWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.ashould_we = qt_to_async(self.should_we, autoresolve=True)

    def should_we(self, stored: StoredUser) -> bool:
        dialog = ShouldWeSaveThisAsDefault(stored, parent=self)
        dialog.exec_()
        return dialog.result() == QtWidgets.QDialog.Accepted

    async def ashould_we_save(self, store: StoredUser) -> bool:
        """Should ask the user if we should save the user"""
        return await self.ashould_we(store)



import os
from typing import Any, Dict, Optional
import pydantic
import datetime
import logging
import json
from .claim import ClaimDemander, ClaimError
from .types import Token
from fakts import Fakts
import logging
from qtpy import QtCore

from fakts.grants.remote.types import FaktsEndpoint

logger = logging.getLogger(__name__)
from typing import List, Dict, Optional

import json
from pydantic import BaseModel, Field
from .types import Token

logger = logging.getLogger(__name__)
from fakts.grants.remote.types import Demander, FaktsEndpoint


class AutoSaveCacheStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    cache_file: str = ".endpoint_cache.json"

    def read_from_cache(self) -> FaktsEndpoint:
        if not os.path.exists(self.cache_file):
            return None

        with open(self.cache_file, "r") as f:
            x = json.load(f)
            try:
                cache = FaktsEndpoint(**x)
                return cache
            except pydantic.ValidationError as e:
                logger.error(f"Could not load cache file: {e}. Ignoring it")
                return None

    def write_to_cache(self, endpoint: Optional[FaktsEndpoint]):
        if endpoint is None:
            os.path.remove(self.cache_file)
            return

        with open(self.cache_file, "w+") as f:
            json.dump(json.loads(endpoint.json()), f)

    async def aput_default_endpoint(self, endpoint: Optional[FaktsEndpoint]) -> None:
        self.write_to_cache(endpoint)

    async def aget_default_endpoint(
        self,
    ) -> Optional[FaktsEndpoint]:
        ...

        return self.read_from_cache()

    class Config:
        arbitrary_types_allowed = True



import os
from typing import Any, Dict, Optional
import pydantic
import datetime
import logging
import json
from .types import Token
from fakts import Fakts
import logging
from qtpy import QtCore

from fakts.grants.remote.types import FaktsEndpoint

logger = logging.getLogger(__name__)
from typing import List, Dict, Optional

import json
from pydantic import BaseModel, Field
from .types import Token

logger = logging.getLogger(__name__)
from fakts.grants.remote.types import Demander, FaktsEndpoint


class EndpointDefaults(BaseModel):
    default_token: Dict[str, Token] = {}


class AutoSaveCacheStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    cache_file: str = ".fakts_cache.json"

    def read_from_cache(self) -> EndpointDefaults:
        if not os.path.exists(self.cache_file):
            with open(self.cache_file, "w+") as f:
                json.dump(EndpointDefaults().json(), f)

        with open(self.cache_file, "r") as f:
            x = json.load(f)
            try:
                cache = EndpointDefaults(**x)
                return cache
            except pydantic.ValidationError as e:
                logger.error(f"Could not load cache file: {e}. Ignoring it")
                return EndpointDefaults()

    def write_to_cache(self, cache: EndpointDefaults):
        with open(self.cache_file, "w+") as f:
            json.dump(json.loads(cache.json()), f)

    async def aput_default_token_for_endpoint(
        self, endpoint: FaktsEndpoint, token: Token
    ) -> None:
        storage = self.read_from_cache()
        if token is None:
            if endpoint.base_url in storage.default_token:
                del storage.default_token[endpoint.base_url]
        else:
            storage.default_token[endpoint.base_url] = token

        self.write_to_cache(storage)

    async def aget_default_token_for_endpoint(
        self, endpoint: FaktsEndpoint
    ) -> Optional[FaktsEndpoint]:
        ...

        storage = self.read_from_cache()
        if endpoint.base_url in storage.default_token:
            return storage.default_token[endpoint.base_url]

        return None

    class Config:
        arbitrary_types_allowed = True



from pydantic import Field
from fakts.grants.base import FaktsGrant
from fakts.grants.errors import GrantError
import ssl
import certifi
from pydantic import BaseModel
import aiohttp
from typing import Any, Dict, Optional, List
from .errors import ClaimError
import logging
from .types import Demander, Discovery, FaktsEndpoint
from fakts.types import FaktsRequest

logger = logging.getLogger(__name__)


Token = str
EndpointUrl = str


class RemoteGrantError(GrantError):
    """Base class for all remotegrant errors"""


class RemoteGrant(FaktsGrant):
    """Abstract base class for remote grants

    A Remote grant is a grant that connects to a fakts server,
    and tires to establishes a secure relationship with it.

    This is done by providing the fakts server with a software
    manifest consisting of a world unique identifier, and a
    version number.

    The fakts server then can depending on the grant type
    respond with a token that then in turn can be used to
    retrieve the configuration from the fakts server.

    """

    discovery: Discovery
    """The discovery mechanism to use for finding the endpoint"""

    demander: Demander
    """The demander mechanism to use for demanding the token FROM the endpoint"""

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    async def aload(self, request: FaktsRequest):
        """Load the configuration from a remote endpoint"""
        endpoint = await self.discovery.discover(request)
        token = await self.demander.ademand(endpoint, request)

        print(endpoint, token)

        return await self.aclaim(token, endpoint)

    async def aclaim(self, token: Token, endpoint: FaktsEndpoint) -> Dict[str, Any]:
        """Claim the configuration from the endpoint"""

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context)
        ) as session:
            async with session.post(
                f"{endpoint.base_url}claim/",
                json={
                    "token": token,
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if not "status" in data:
                        raise ClaimError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise ClaimError(data["message"])
                    if status == "granted":
                        return data["config"]

                    raise ClaimError(f"Unexpected status: {status}")
                else:
                    raise Exception("Error! Coud not claim this app on this endpoint")

    class Config:
        arbitrary_types_allowed = True



from fakts import Fakts
import logging
from qtpy import QtCore

from fakts.grants.remote.types import FaktsEndpoint

logger = logging.getLogger(__name__)
from typing import List, Dict, Optional

import json
from pydantic import BaseModel, Field
from fakts.grants.remote.demanders.types import Token


class EndpointDefaults(BaseModel):
    default_token: Dict[str, Token] = {}


class AutoSaveTokenStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    settings: QtCore.QSettings
    save_key: str

    async def aput_default_token_for_endpoint(
        self, endpoint: FaktsEndpoint, token: Token
    ) -> None:
        un_storage = self.settings.value(self.save_key, None)
        if not un_storage:
            storage = EndpointDefaults()
        else:
            try:
                storage = EndpointDefaults(**json.loads(un_storage))
            except Exception as e:
                storage = EndpointDefaults()

        if token is None:
            if endpoint.base_url in storage.default_token:
                del storage.default_token[endpoint.base_url]
        else:
            storage.default_token[endpoint.base_url] = token

        self.settings.setValue(self.save_key, storage.json())

    async def aget_default_token_for_endpoint(
        self, endpoint: FaktsEndpoint
    ) -> Optional[FaktsEndpoint]:
        ...

        un_storage = self.settings.value(self.save_key, None)
        if not un_storage:
            return None
        try:
            storage = EndpointDefaults(**json.loads(un_storage))
            if endpoint.base_url in storage.default_token:
                return storage.default_token[endpoint.base_url]
        except Exception as e:
            print(e)
            return None

        return None

    class Config:
        arbitrary_types_allowed = True



from fakts.grants.remote.discovery.base import Discovery
from fakts.grants.remote.discovery.base import FaktsEndpoint
from fakts.grants.remote.discovery.errors import DiscoveryError
import ssl
import certifi
from pydantic import Field
import aiohttp
import logging
from typing import List
import asyncio
from .utils import discover_url

logger = logging.getLogger(__name__)


class WellKnownDiscovery(Discovery):
    url = "http://localhost:8000"
    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    allow_appending_slash: bool = Field(
        default=True,
        description="If the url does not end with a slash, should we append one? ",
    )
    auto_protocols: List[str] = Field(
        default_factory=lambda: [],
        description="If no protocol is specified, we will try to connect to the following protocols",
    )
    timeout: int = Field(
        default=3,
        description="The timeout for the connection",
    )

    async def discover(self, request):
        return await discover_url(
            self.url,
            self.ssl_context,
            auto_protocols=self.auto_protocols,
            allow_appending_slash=self.allow_appending_slash,
            timeout=self.timeout,
        )

    class Config:
        extra = "forbid"
        arbitrary_types_allowed = True



from fakts.grants.base import FaktsGrant
import yaml
from fakts.types import FaktsRequest


class YamlGrant(FaktsGrant):
    filepath: str

    async def aload(self, request: FaktsRequest):
        with open(self.filepath, "r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        return config



from typing import Protocol, runtime_checkable, Dict, Any, Optional
from pydantic import BaseModel
from fakts.types import FaktsRequest

Token = str


class FaktsEndpoint(BaseModel):
    base_url = "http://localhost:8000/f/"
    name: str = "Helper"
    description: Optional[str]
    retrieve_url: Optional[str]
    claim_url: Optional[str]
    version: Optional[str]


@runtime_checkable
class Demander(Protocol):
    """A demander takes a FaktsEndpoint and returns the Fakts
    user input.
    """

    async def ademand(self, endpoint: FaktsEndpoint, request: FaktsRequest) -> Token:
        ...


@runtime_checkable
class Discovery(Protocol):
    """Discovery is the abstract base class for discovery mechanisms

    A discovery mechanism is a way to find a Fakts endpoint
    that can be used to retrieve the configuration.

    This class provides an asynchronous interface, as the discovery can
    envolve lenghty operations such as network requests or waiting for
    user input.
    """

    async def discover(self, request: FaktsRequest) -> FaktsEndpoint:
        ...



from fakts.grants.remote.discovery.base import FaktsEndpoint
import aiohttp
import logging
import ssl
from fakts.grants.remote.discovery.errors import DiscoveryError
from typing import Optional, List

logger = logging.getLogger(__name__)


async def check_wellknown(
    url: str, ssl_context: ssl.SSLContext, timeout=4
) -> FaktsEndpoint:
    url = f"{url}.well-known/fakts"

    async with aiohttp.ClientSession(
        connector=aiohttp.TCPConnector(ssl=ssl_context),
        headers={"User-Agent": "Fakts/0.1", "Accept": "application/json"},
    ) as session:
        async with session.get(
            url,
            timeout=timeout,
        ) as resp:
            if resp.status == 200:
                data = await resp.json()

                if "name" not in data:
                    logger.error(f"Malformed answer: {data}")
                    raise Exception("Malformed Answer")

                return FaktsEndpoint(**data)

            else:
                logger.error(f"Could not retrieve on the endpoint: {resp.status}")
                raise DiscoveryError(
                    f"Error! We could not retrieve the endpoint. {url} "
                )


async def discover_url(
    url: str,
    ssl_context: ssl.SSLContext,
    auto_protocols: Optional[List[str]] = None,
    allow_appending_slash: bool = False,
    timeout=4,
) -> FaktsEndpoint:
    if "://" not in url:
        logger.info(f"No protocol specified on {url}")
        if not auto_protocols or len(auto_protocols) == 0:
            raise DiscoveryError(
                "No protocol specified and no auto protocols specified"
            )

        errors = []

        for protocol in auto_protocols:
            logger.info(f"Trying to connect to {protocol}://{url}")
            try:
                if allow_appending_slash and not url.endswith("/"):
                    url = f"{url}/"

                return await check_wellknown(
                    f"{protocol}://{url}", ssl_context, timeout=timeout
                )
            except Exception as e:
                logger.info(f"Could not connect to {protocol}://{url}")
                errors.append((protocol, e))
                continue

        errors_string = "\n".join(
            [f"- {protocol}://{url}\n  " + str(e) for protocol, e in errors]
        )

        raise DiscoveryError(f"Could not connect via any protocol: \n{errors_string}")

    if allow_appending_slash and not url.endswith("/"):
        url = f"{url}/"

    return await check_wellknown(url, ssl_context, timeout=timeout)



from fakts.grants.remote.discovery.beacon import retrieve_bindings, advertise
from fakts.grants.remote.discovery.base import Beacon
import rich_click as click
import asyncio
from rich.console import Console
from rich.prompt import Prompt
from rich.panel import Panel
import os
import sys

console = Console()

logo = """
    __      _   _      
  / _|__ _| |_| |_ ___
 |  _/ _` | / /  _(_-/
 |_| \__,_|_\_\\__/__/                     
 """

welcome = "The fakts beacon lets you advertise a fakts endpoint on your local network. This is useful if you want to advertise a fakts endpoint to other devices on your local network."


async def adversite_all(bindings, url, interval=1, iterations=10):
    await asyncio.gather(
        *[
            advertise(
                binding, [Beacon(url=url)], interval=interval, iterations=iterations
            )
            for binding in bindings
        ]
    )


def advertise_cli(url: str, all=None, interval=1, iterations=10):
    """Advertises the given endpoint on the interface specified by the user


    Args:
        name (str, optional): The name of the faktsendpoint. Defaults to None (user can specify it)
        url (str, optional): The url of the faktsendpont. Defaults to None (user can specify it)
    """


@click.group()
def cli():
    """
    Fakts cli lets you interact with the fakts api
    through the command line.

    """
    pass


@cli.command("beacon", short_help="Advertises a fakts endpoint")
@click.argument("url")
@click.option("--all", "-a", help="Advertise on all interfaces", is_flag=True)
@click.option(
    "--iterations",
    "-i",
    help="How many iterations (-1 equals indefintetly)",
    default=-1,
)
@click.option(
    "--interval", "-i", help="Which second interval between broadcast", default=5
)
def beacon(url, all, iterations, interval):
    """Runs the arkitekt app (using a builder)"""

    md = Panel(logo + welcome, title="Fakts Beacon", title_align="center")
    console.print(md)
    try:
        bindings = retrieve_bindings()
    except ImportError as e:
        error = Panel(
            """netifaces is required to use the advertised discovery. please install it seperately or install fakts with the 'beacon' extras: pip install "fakts\[beacon]" """,
            title="Fakts Beacon",
            title_align="center",
            style="red",
        )
        console.print(error)
        sys.exit(1)

    console.print("Which Interface should be used for broadcasting?")

    all = (
        all
        or Prompt.ask(
            "Do you want to use all interfaces?", default="y", choices=["y", "n"]
        )
        == "y"
    )

    console.print(
        f"Advertising endpoint every {interval} seconds " + "forever"
        if iterations == -1
        else f"{iterations} times"
    )

    if not all:
        for i, binding in enumerate(bindings):
            console.print(
                f"[{i}] : Use interface {binding.interface}: {binding.bind_addr} advertising to {binding.broadcast_addr}"
            )

        bind_index = Prompt.ask(
            "Which Binding do you want to use?",
            default=1,
            choices=[str(i) for i in range(len(bindings))],
        )

        with console.status("Advertising beacon") as status:
            asyncio.run(
                advertise(
                    bindings[int(bind_index)],
                    [Beacon(url=url)],
                    interval=interval,
                    iterations=iterations,
                )
            )

    else:
        with console.status("Advertising beacons") as status:
            asyncio.run(
                adversite_all(bindings, url, interval=interval, iterations=iterations)
            )


if __name__ == "__main__":
    cli()



import json
from typing import runtime_checkable, Protocol
from herre.grants.stored_login import (
    UserStore,
    StoredUser,
)
from qtpy import QtWidgets, QtCore
from koil.qt import qt_to_async
import token
from fakts.grants.remote.demanders.types import Token
from fakts.grants.remote.types import FaktsEndpoint


class ShouldWeSaveThisAsDefault(QtWidgets.QDialog):
    def __init__(self, endpoint: FaktsEndpoint, token: Token, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.setWindowTitle(f"Connected to {endpoint.name}")

        self.qlabel = QtWidgets.QLabel(
            f"Do you want to auto save this configuration for this endpoint?"
        )

        vlayout = QtWidgets.QVBoxLayout()
        self.setLayout(vlayout)

        vlayout.addWidget(self.qlabel)

        hlayout = QtWidgets.QHBoxLayout()
        vlayout.addLayout(hlayout)

        self.yes_button = QtWidgets.QPushButton("Yes")
        self.no_button = QtWidgets.QPushButton("No")

        self.yes_button.clicked.connect(self.on_yes)
        self.no_button.clicked.connect(self.on_no)

        hlayout.addWidget(self.yes_button)
        hlayout.addWidget(self.no_button)

    def on_yes(self):
        self.accept()

    def on_no(self):
        self.reject()


class AutoSaveTokenWidget(QtWidgets.QWidget):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.ashould_we = qt_to_async(self.should_we, autoresolve=True)

    def should_we(self, endpoint: FaktsEndpoint, token: Token) -> bool:
        dialog = ShouldWeSaveThisAsDefault(endpoint, token, parent=self)
        dialog.exec_()
        return dialog.result() == QtWidgets.QDialog.Accepted

    async def ashould_we_save(self, endpoint: FaktsEndpoint, token: Token) -> bool:
        """Should ask the user if we should save the user"""
        return await self.ashould_we(endpoint, token)



from fakts import Fakts
import logging
from qtpy import QtCore

from fakts.grants.remote.types import FaktsEndpoint

logger = logging.getLogger(__name__)
from typing import List, Dict, Optional

import json
from pydantic import BaseModel, Field


class AutoSaveEndpointStore(BaseModel):
    """Retrieves and stores users matching the currently
    active fakts grant"""

    settings: QtCore.QSettings
    save_key: str

    async def aput_default_endpoint(self, endpoint: Optional[FaktsEndpoint]) -> None:
        self.settings.setValue(self.save_key, endpoint.json() if endpoint else None)

    async def aget_default_endpoint(self) -> Optional[FaktsEndpoint]:
        ...

        un_storage = self.settings.value(self.save_key, None)
        if not un_storage:
            return None
        try:
            storage = FaktsEndpoint.parse_raw(un_storage)
            return storage
        except Exception as e:
            print(e)

        return None

    class Config:
        arbitrary_types_allowed = True



from pydantic import BaseModel
from fakts.types import FaktsRequest


class FaktsGrant(BaseModel):
    """Abstract Base Class for Fakts Grants

    Grants are used to load the configuration for Fakts. They are used to
    load configuration from a local file, a remote endpoint, a database, or
    any other source.

    A grant needs to implement the `aload` function, which is an
    async function. This means that it should not depend on cpu bound tasks.
    If you need to do cpu bound tasks, you should use a thread pool executor
    to do so.

    """

    async def aload(self, request: FaktsRequest):
        raise NotImplementedError("Fakts need to implement this function")



from typing import List
from fakts.grants.base import FaktsGrant
import asyncio
from functools import reduce

from fakts.utils import update_nested
from fakts.types import FaktsRequest


class ParallelGrant(FaktsGrant):
    """A grant that loads multiple grants in parallel and merges the results"""

    grants: List[FaktsGrant]
    omit_exceptions = False
    " Omit exceptions if any of the grants fail to load "

    async def aload(self, request: FaktsRequest):
        config_futures = [grant.aload(request) for grant in self.grants]
        configs = await asyncio.gather(
            config_futures, return_exceptions=self.omit_exceptions
        )
        return reduce(
            lambda x, y: update_nested(x, y) if isinstance(y, dict) else x, configs, {}
        )



from .env import EnvGrant
from .base import FaktsGrant
from .errors import GrantError
from .meta import FailsafeGrant, ParallelGrant, CacheGrant


__all__ = [
    "EnvGrant",
    "GrantError",
    "FaktsGrant",
    "YamlGrant",
    "FailsafeGrant",
    "ParallelGrant",
    "CacheGrant",
]



from typing import Any, Dict, List, Optional, Type, runtime_checkable, Protocol
from pydantic import BaseModel, Field, SecretStr

import ssl
import certifi
import aiohttp
import logging
import asyncio
from fakts.grants.remote.types import FaktsEndpoint
from herre.fetcher.types import UserFetcher
from fakts.grants.remote.types import Discovery
from fakts.types import FaktsRequest


logger = logging.getLogger(__name__)


@runtime_checkable
class EndpointStore(Protocol):
    async def aget_default_endpoint(self) -> Optional[FaktsEndpoint]:
        ...

    async def aput_default_endpoint(self, endpoint: FaktsEndpoint) -> None:
        ...


@runtime_checkable
class AutoSaveDecider(Protocol):
    async def ashould_we_save(self, endpoint: FaktsEndpoint) -> bool:
        """Should ask the user if he wants to save the endpoint"""
        ...


class StaticDecider(BaseModel):
    allow_save: bool = True

    async def ashould_we_save(self, *args, **kwargs) -> bool:
        return self.allow_save


class AutoSaveDiscovery(BaseModel):
    """A discovery the autosaves the
    discovered endpoint and selects it as the default one.



    """

    store: EndpointStore
    """this is the login widget (protocol)"""

    decider: AutoSaveDecider = Field(default_factory=lambda: StaticDecider())
    """this is the login widget (protocol)"""

    discovery: Discovery
    """The grant to use for the login flow."""

    async def discover(self, request: FaktsRequest) -> FaktsEndpoint:
        """Fetches the token

        This function will only delegate to the grant if the user has not
        previously logged in (aka there is no token in the storage) Or if the
        force_refresh flag is set.

        Args:
            force_refresh (bool, optional): _description_. Defaults to False.

        Raises:
            e: _description_

        Returns:
            Token: _description_
        """

        try:
            if request.context.get("allow_auto_discover", True):
                stored_endpoint: EndpointStore = (
                    await self.store.aget_default_endpoint()
                )
                if stored_endpoint:
                    print("Discovered stored enpoint", stored_endpoint)
                    # Lets check if the token is still valid
                    return stored_endpoint

                    # This time with a refresh

            # We are skipping the widget and just fetching the token

            print("Discovering endpoint", request)
            endpoint = await self.discovery.discover(request)
            print("Discovered endpoint", endpoint)
            should_we_save = await self.decider.ashould_we_save(endpoint)
            if should_we_save:
                await self.store.aput_default_endpoint(endpoint)
            else:
                await self.store.aput_default_endpoint(None)

            return endpoint

        except asyncio.CancelledError as e:
            raise e

        except Exception as e:
            logger.error(e, exc_info=True)
            raise e

    class Config:
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True



import aiohttp
from typing import Optional
from pydantic import Field
from .types import Token
from .errors import DemandError
from pydantic import BaseModel, Field
import logging
from fakts.grants.remote.types import FaktsEndpoint, FaktsRequest
import ssl
import certifi

logger = logging.getLogger(__name__)


class RetrieveError(DemandError):
    pass


class RetrieveDemander(BaseModel):
    """Retrieve Demander

    A retrieve grant is a remote grant can be used to retrieve a token and a configuration from a fakts server, by claiming to be an already
    registed public application on the fakts server. Public applications are applications that are not able to keep a secret, and therefore
    need users to explicitly grant them access to their data. YOu need to also provide a redirect_uri that matches the one that is registered
    on the fakts server.

    """

    ssl_context: ssl.SSLContext = Field(
        default_factory=lambda: ssl.create_default_context(cafile=certifi.where()),
        exclude=True,
    )
    """ An ssl context to use for the connection to the endpoint"""

    manifest: BaseModel
    retrieve_url: Optional[str] = Field(
        None,
        description="The url to use for retrieving the token (overwrited the endpoint url)",
    )

    async def ademand(self, endpoint: FaktsEndpoint, request: FaktsRequest) -> Token:
        retrieve_url = (
            self.retrieve_url
            or endpoint.retrieve_url
            or f"{endpoint.base_url}retrieve/"
        )

        async with aiohttp.ClientSession(
            connector=aiohttp.TCPConnector(ssl=self.ssl_context)
        ) as session:
            logger.debug(f"Requesting token from {retrieve_url}")
            async with session.post(
                retrieve_url,
                json={
                    "manifest": self.manifest.dict(),
                },
            ) as resp:
                data = await resp.json()

                if resp.status == 200:
                    data = await resp.json()
                    if not "status" in data:
                        raise RetrieveError("Malformed Answer")

                    status = data["status"]
                    if status == "error":
                        raise RetrieveError(data["message"])
                    if status == "granted":
                        return data["token"]

                    raise RetrieveError(f"Unexpected status: {status}")
                else:
                    raise Exception("Error! Coud not claim this app on this endpoint")

    class Config:
        arbitrary_types_allowed = True
